 /*
    Peter's Picross - A remake for TI 68k-based calculators of the Nintendo Gameboy game, "Mario's Picross"
    Copyright (C) 2008  Peter Fernandes
    supersonicandtails@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    
    
    engine.c: in-game play and puzzle editing
*/

#include <tigcclib.h>
#include "main.h"
#include "engine.h"
#include "gfx.h"
#include "editor.h"
#include "files.h"


void draw_row_quadrupled(short x, short y, short row, short width) {
	int i=width-1;
	int pos=x;
	
	
	for(;i>=0;i--,pos+=4) {
		if(row&(1<<i)) {
			DrawLine(pos,y,pos+3,y,A_NORMAL);
		}
	}
}

void seconds2str(short seconds, char *buf) {
	sprintf(buf,"%d%d:%d%d",(seconds/600),(seconds/60)%10,(seconds-((seconds/600)*600)-((seconds/60)%10)*60)/10,(seconds%10));
}

//returns where to draw the vertical guideline to
short line_boundary(PUZZLE *puzzle, short offset) {
	if((puzzle->maxnumy+(puzzle->dim*6))>LCD_HEIGHT-1) {
		if(offset==(15-((LCD_HEIGHT-puzzle->maxnumy)/6)))
			return LCD_HEIGHT-5;
		
		return LCD_HEIGHT-1;
	}
		
		
	return (puzzle->maxnumy+(puzzle->dim*6));
}


short main_loop(PUZZLE *puzzle, short mode, short *buf) {
	char solved=0;
	unsigned short key;
	void *kbq=kbd_queue();
	short x;
	short y;
	short oldx;
	short oldy;
	short posx=0;
	short posy=0;
	short offset=0;
	unsigned short board_cleared[puzzle->dim];
	unsigned short board_marked[puzzle->dim];
	char justchanged=0;
	short seconds=1800;
	char bads=0;
	char timestr[6];
	short maxposy;
	char err_sec=0;
	char *screen=NULL;
	short newmode=mode;

	
	ClrScr();
	
	
	FontSetSys(F_4x6);
	
	
	memset(board_cleared,0,sizeof(board_cleared));
	memset(board_marked,0,sizeof(board_marked));
	
	
	if(mode==MODE_RESTORE) {
		//filename of puzzle (9), seconds (2), board cleared (15), board marked (15), padding (1)
		screen=settings_file+30;
		
		seconds=*(((short *)screen)++);
		
		memcpy(board_cleared,screen,15*sizeof(unsigned short));
		screen+=15*sizeof(unsigned short);
		memcpy(board_marked,screen,15*sizeof(unsigned short));
		
		newmode=MODE_PLAY;
	}
	
	if(!(screen=malloc(LCD_SIZE))) {
		return -1;
	}
	
	if(newmode==MODE_PLAY) {
		calc_numbers(puzzle);
		
		y=puzzle->maxnumy+1;
		oldy=puzzle->maxnumy+1;
		x=puzzle->maxnumx+2;
		oldx=puzzle->maxnumx+2;
	}
	else {
		puzzle->maxnumy=1;
		puzzle->maxnumx=0;
		x=2;
		y=2;
		oldx=2;
		oldy=2;
	}
	
	#ifdef DEBUG1
		solved=1;
		goto bottom;
	#endif
		
		
	if(newmode!=MODE_EDIT)
		DrawStr(LCD_WIDTH-18,0,"Time:",A_NORMAL);
		
	
	update_board(puzzle,board_cleared,newmode==MODE_PLAY?board_marked:NULL,-1);
	
	maxposy=(puzzle->maxnumy+(puzzle->dim*6))<LCD_HEIGHT?puzzle->dim-1:((LCD_HEIGHT-puzzle->maxnumy)/6)-1;
	
	Sprite8(x,y,5,tile_cursor,LCD_MEM,SPRT_XOR);
	
	
	
	//draw scroll bar
	if(((LCD_HEIGHT-puzzle->maxnumy)/6)<puzzle->dim) {
		DrawClipRect(&(WIN_RECT){puzzle->maxnumx+(puzzle->dim*6)+3,puzzle->maxnumy+6,puzzle->maxnumx+(puzzle->dim*6)+7,LCD_HEIGHT-7},&(SCR_RECT){{puzzle->maxnumx+(puzzle->dim*6)+3,puzzle->maxnumy+6,puzzle->maxnumx+(puzzle->dim*6)+7,LCD_HEIGHT-7}},A_NORMAL);
		Sprite8(puzzle->maxnumx+(puzzle->dim*6)+3,puzzle->maxnumy,6,arrow_up,LCD_MEM,SPRT_OR);
		Sprite8(puzzle->maxnumx+(puzzle->dim*6)+3,LCD_HEIGHT-6,6,arrow_down,LCD_MEM,SPRT_OR);
	}
	
	//set highlight bars
	
	if(newmode!=MODE_EDIT) {
		ScrRectFill(&(SCR_RECT){{0,y,puzzle->maxnumx-1,y+4}},&(SCR_RECT){{0,y,puzzle->maxnumx-1,y+4}},A_XOR);
		ScrRectFill(&(SCR_RECT){{x,0,x+4,puzzle->maxnumy-2}},&(SCR_RECT){{x,0,x+4,puzzle->maxnumy-2}},A_XOR);
	}

	OSFreeTimer(USER_TIMER);
	OSRegisterTimer(USER_TIMER,20);
	
	while(!solved) {
	
		//if the cursor moved, update it
		if(oldx!=x || oldy!=y) {
			if(!justchanged) {
				Sprite8(oldx,oldy,5,tile_cursor,LCD_MEM,SPRT_XOR);
			}
			justchanged=0;
			
			
			Sprite8(x,y,5,tile_cursor,LCD_MEM,SPRT_XOR);
		}
		
		oldx=x;
		oldy=y;
		
		
		//save bottom of screen from 2ND/Alpha
		memcpy(screen,LCD_MEM+(LCD_HEIGHT*30)-(5*30),5*30);
		
		
		if(puzzle->dim==15) {
			DrawLine(puzzle->maxnumx+30,puzzle->maxnumy,puzzle->maxnumx+30,line_boundary(puzzle,offset),A_NORMAL);
			DrawLine(puzzle->maxnumx+60,puzzle->maxnumy,puzzle->maxnumx+60,line_boundary(puzzle,offset),A_NORMAL);
			
			if((puzzle->maxnumy+29-(offset*6))>puzzle->maxnumy)
				DrawLine(puzzle->maxnumx+1,puzzle->maxnumy+29-(offset*6),puzzle->maxnumx+(puzzle->dim*6),puzzle->maxnumy+29-(offset*6),A_NORMAL);
			
			DrawLine(puzzle->maxnumx+1,puzzle->maxnumy+59-(offset*6),puzzle->maxnumx+(puzzle->dim*6),puzzle->maxnumy+59-(offset*6),A_NORMAL);
		}
		
		
		//draw preview
		Sprite16(LCD_WIDTH-1-puzzle->dim,LCD_HEIGHT-1-puzzle->dim,puzzle->dim,board_blank,LCD_MEM,SPRT_AND);
		Sprite16(LCD_WIDTH-1-puzzle->dim,LCD_HEIGHT-1-puzzle->dim,puzzle->dim,board_cleared,LCD_MEM,SPRT_OR);
		
		
		do {
			if(newmode!=MODE_EDIT) {
				if(OSTimerExpired(USER_TIMER) || seconds==1800) {
					OSTimerRestart(USER_TIMER);
					seconds2str(seconds,timestr);
					DrawStr(LCD_WIDTH-18,6,timestr,A_REPLACE);
					seconds--;
					
					if(!err_sec)
						ScrRectFill(&(SCR_RECT){{LCD_WIDTH-18,13,LCD_WIDTH-1,63}},&(SCR_RECT){{LCD_WIDTH-18,13,LCD_WIDTH-1,63}},A_REVERSE);
					
					if(err_sec>0)
						err_sec--;
				}
				
				
				if(seconds<=0) {
					solved=-1;
					goto bottom;
				}
			}
			
			
			
			
		} while(OSdequeue(&key,kbq));
		
				
		//restore status line area of board
		memcpy(LCD_MEM+(LCD_HEIGHT*30)-(5*30),screen,5*30);
		
		
		if(newmode!=MODE_EDIT) {
			if((key&~KB_AUTOREPEAT)==KEY_UP || (key&~KB_AUTOREPEAT)==KEY_DOWN)
				ScrRectFill(&(SCR_RECT){{0,y,puzzle->maxnumx-1,y+4}},&(SCR_RECT){{0,y,puzzle->maxnumx-1,y+4}},A_XOR);
			else if((key&~KB_AUTOREPEAT)==KEY_LEFT || (key&~KB_AUTOREPEAT)==KEY_RIGHT)
				ScrRectFill(&(SCR_RECT){{x,0,x+4,puzzle->maxnumy-2}},&(SCR_RECT){{x,0,x+4,puzzle->maxnumy-2}},A_XOR);
			
		}

		if((key&~KB_AUTOREPEAT)==KEY_UP) {
			if(posy>offset) {
				y-=6;
				posy--;
			}
			else if(posy > 0) { //scroll the puzzle
				posy--;
				offset--;
				update_board(puzzle,board_cleared,newmode==MODE_PLAY?board_marked:NULL,offset);
				Sprite8(x,y,5,tile_cursor,LCD_MEM,SPRT_XOR);
			}
		}
		else if((key&~KB_AUTOREPEAT)==KEY_DOWN) {
			if(posy<maxposy || posy-offset<maxposy) {
				y+=6;
				posy++;
			}
			else if(posy==maxposy+offset && posy<puzzle->dim-1) { //scroll the puzzle
				posy++;
				offset++;
				update_board(puzzle,board_cleared,newmode==MODE_PLAY?board_marked:NULL,offset);
				
				Sprite8(x,y,5,tile_cursor,LCD_MEM,SPRT_XOR);
			}
		}
		else if((key&~KB_AUTOREPEAT)==KEY_LEFT) {
			if(posx) {
				posx--;
				x-=6;
			}
		}
		else if((key&~KB_AUTOREPEAT)==KEY_RIGHT) {
			if(posx<puzzle->dim-1) {
				posx++;
				x+=6;
			}
		}
		else if((key&~KB_AUTOREPEAT)==KEY_F1) {
			board_marked[posy]&=~(1<<((puzzle->dim)-1-posx));

			if(newmode==MODE_EDIT || (puzzle->board[posy]&(1<<((puzzle->dim)-1-posx)))) {
				
				board_cleared[posy]^=1<<((puzzle->dim)-1-posx);
				
				
				//clear the square
				if((board_cleared[posy]&(1<<((puzzle->dim)-1-posx)))) {
					Sprite8(x,y,5,tile_mask,LCD_MEM,SPRT_AND);
					Sprite8(x+1,y+1,5,tile_cursor,LCD_MEM,SPRT_OR);
					Sprite8(x,y,5,tile_cursor,LCD_MEM,SPRT_XOR);
				}
				else { //re-bury the square
					Sprite8(x,y,5,tile_normal_mask,LCD_MEM,SPRT_AND);
					Sprite8(x,y,5,tile_cursor,LCD_MEM,SPRT_XOR);
				}
				
				justchanged=0;
				
				if(newmode!=MODE_EDIT)
					solved=check_solved(puzzle,board_cleared);
			}
			else { //subtract time from the clock
				seconds-=120+(bads*120)+(bads>1?120:0);
									
				if(seconds<0)
					seconds=0;
					
				err_sec=2;
										
				//update the clock
				seconds2str(seconds,timestr);
				DrawStr(LCD_WIDTH-18,6,timestr,A_REPLACE);
				
				sprintf(timestr,"-%d:%d%d",((120+(bads*120)+(bads>1?120:0))/60)%10,((120+(bads*120)+(bads>1?120:0))-(((120+(bads*120)+(bads>1?120:0))/600)*600)-(((120+(bads*120)+(bads>1?120:0))/60)%10)*60)/10,((120+(bads*120)+(bads>1?120:0))%10));
				DrawStr(LCD_WIDTH-18,13,timestr,A_NORMAL);
				
				FontSetSys(F_8x10);
				DrawChar(LCD_WIDTH-9,25,'M',A_NORMAL);
				DrawChar(LCD_WIDTH-9,35,'I',A_NORMAL);
				DrawChar(LCD_WIDTH-9,45,'S',A_NORMAL);
				DrawChar(LCD_WIDTH-9,55,'S',A_NORMAL);
				FontSetSys(F_4x6);
				
				if(bads<2)
					bads++;
			}
		}
		else if((key&~KB_AUTOREPEAT)==KEY_F2) {
			if(newmode!=MODE_EDIT) { //no need to mark tiles when editing
				board_marked[posy]^=(1<<((puzzle->dim)-1-posx));
				
				board_cleared[posy]&=~(1<<((puzzle->dim)-1-posx));
				
				if(board_marked[posy]&(1<<((puzzle->dim)-1-posx))) {
					Sprite8(x,y,5,tile_mask,LCD_MEM,SPRT_AND);
					Sprite8(x+1,y+1,3,tile_marked,LCD_MEM,SPRT_OR);
					Sprite8(x,y,5,tile_cursor,LCD_MEM,SPRT_XOR);
					justchanged=0;
				}
				else {
					Sprite8(x,y,5,tile_mask,LCD_MEM,SPRT_AND);
					Sprite8(x,y,5,tile_cursor,LCD_MEM,SPRT_OR);

					justchanged=0;
				}
			}
		}
		else if((key&~KB_AUTOREPEAT)==KEY_ENTER) {
			if(newmode==MODE_EDIT) { //finished editing
				FontSetSys(F_4x6);
				DrawStr(LCD_WIDTH-35,LCD_HEIGHT-18,"DONE?",A_REPLACE);
				DrawStr(LCD_WIDTH-45,LCD_HEIGHT-9,"(Enter/ESC)",A_REPLACE);
				
				FontSetSys(F_6x8);
				
				while((key=ngetchx())!=KEY_ESC) {
					if(key==KEY_ENTER)
						goto bottom;
				}
				
				ScrRectFill(&(SCR_RECT){{LCD_WIDTH-45,LCD_HEIGHT-18,LCD_WIDTH-1,LCD_HEIGHT-1}},&(SCR_RECT){{LCD_WIDTH-45,LCD_HEIGHT-18,LCD_WIDTH-1,LCD_HEIGHT-1}},A_REVERSE);

				
				
			}
			else if(screen) { //Pause
				memcpy(screen,LCD_MEM,LCD_SIZE);
				ClrScr();
				
				FontSetSys(F_6x8);
				DrawStr((LCD_WIDTH/2)-(15*3),0,"Peter's Picross",A_NORMAL);
				DrawStr((LCD_WIDTH/2)-(6*3),(LCD_HEIGHT/2)-4,"PAUSED",A_NORMAL);
				FontSetSys(F_4x6);
				
				key=0;
				
				do {
					OSdequeue(&key,kbq);
					
				} while(key!=KEY_ENTER);
				
				memcpy(LCD_MEM,screen,LCD_SIZE);
			}
		}
		else if((key&~KB_AUTOREPEAT)==KEY_F5) {
			FontSetSys(F_6x8);
			if(screen)
				free(screen);
				
			//save game state
			
			//filename of puzzle (9),padding(1) seconds (2), board cleared (15), board marked (15)
			
			if(mode!=MODE_EDIT) {
				SAVE_STATE
			}
			
			return -1;
		}
		else if((key&~KB_AUTOREPEAT)==KEY_ESC) {
			memcpy(screen,LCD_MEM,LCD_SIZE);
			ClrScr();
			
			FontSetSys(F_4x6);
			
			DrawStr(LCD_WIDTH-35,LCD_HEIGHT-18,"Quit?",A_NORMAL);
			DrawStr(LCD_WIDTH-45,LCD_HEIGHT-9,"(Enter/ESC)",A_NORMAL);
			
			
			while((key=ngetchx())!=KEY_ESC) {
				if(key==KEY_ENTER) {
					if(mode==MODE_EDIT) {
						return -1;
					}
					else {
						if(screen)
							free(screen);
						
						SAVE_STATE
						return 0;
					}
				}
			}
			
			memcpy(LCD_MEM,screen,LCD_SIZE);
		}
		
		
		
		if(newmode!=MODE_EDIT) {
			if((key&~KB_AUTOREPEAT)==KEY_UP || (key&~KB_AUTOREPEAT)==KEY_DOWN)
				ScrRectFill(&(SCR_RECT){{0,y,puzzle->maxnumx-1,y+4}},&(SCR_RECT){{0,y,puzzle->maxnumx-1,y+4}},A_XOR);
			else if((key&~KB_AUTOREPEAT)==KEY_LEFT || (key&~KB_AUTOREPEAT)==KEY_RIGHT)
				ScrRectFill(&(SCR_RECT){{x,0,x+4,puzzle->maxnumy-2}},&(SCR_RECT){{x,0,x+4,puzzle->maxnumy-2}},A_XOR);
		}
		

	}
	
	
	bottom:
	
	if(screen)
		free(screen);
	
	FontSetSys(F_6x8);
	
	if(newmode==MODE_PLAY) {
		ClrScr();

		if(solved==1) {
			DrawStr((LCD_WIDTH/2)-(7*6),0,"Puzzle Solved!",A_NORMAL);
			
			OSFreeTimer(USER_TIMER);
			OSRegisterTimer(USER_TIMER,1);
			
			//draw the picture frame
			DrawClipRect(&(WIN_RECT){((LCD_WIDTH-(puzzle->dim*4))/2)-2,21,((LCD_WIDTH-(puzzle->dim*4))/2)+(puzzle->dim*4)+2,(25+(puzzle->dim*4))},&(SCR_RECT){{((LCD_WIDTH-(puzzle->dim*4))/2)-2,21,((LCD_WIDTH-(puzzle->dim*4))/2)+(puzzle->dim*4)+2,(25+(puzzle->dim*4))}},A_NORMAL);
			DrawLine(((LCD_WIDTH-(puzzle->dim*4))/2)-2,(24+puzzle->dim*4),((LCD_WIDTH-(puzzle->dim*4))/2)+(puzzle->dim*4)+1,(24+(puzzle->dim*4)),A_NORMAL);
			DrawLine(((LCD_WIDTH-(puzzle->dim*4))/2)+(puzzle->dim*4)+1,22,((LCD_WIDTH-(puzzle->dim*4))/2)+(puzzle->dim*4)+1,(24+puzzle->dim*4),A_NORMAL);
			
			//draw the answer picture
			for(newmode=0;newmode<puzzle->dim*4;newmode++) {
				draw_row_quadrupled((LCD_WIDTH-(puzzle->dim*4))/2,23+newmode,puzzle->board[newmode/4],puzzle->dim);
				
				while(!OSTimerExpired(USER_TIMER));
				OSTimerRestart(USER_TIMER);
			}
			
			FontSetSys(F_4x6);
			DrawStr((LCD_WIDTH-DrawStrWidth(puzzle->name,F_4x6))/2,30+newmode,puzzle->name,A_NORMAL);
			
			OSFreeTimer(USER_TIMER);
			OSRegisterTimer(USER_TIMER,20);
			while(!OSTimerExpired(USER_TIMER));
			
			OSFreeTimer(USER_TIMER);
			
			screen=LCD_MEM;
			
			if((((LCD_WIDTH-(puzzle->dim*4))/2)-5)<((LCD_WIDTH-DrawStrWidth(puzzle->name,F_4x6))/2))
				offset=0;
			else
				offset=1;
			
			//scroll the whole thing over to the right
			for(bads=0;bads<(offset?((LCD_WIDTH-DrawStrWidth(puzzle->name,F_4x6))/2):(((LCD_WIDTH-(puzzle->dim*4))/2)-5));bads++) {
				for(posy=21;posy<99;posy++) {
					for(posx=0;posx<20;posx++) {
						*(screen+(posy*30)+posx)<<=1;
						if(posx<99) {
							*(screen+(posy*30)+posx)|=!!(*(screen+(posy*30)+posx+1)&0b10000000);
						}
					}
				}
			}
			
			//draw stats
			DrawStr(9+(offset?DrawStrWidth(puzzle->name,F_4x6):(puzzle->dim*4)),21,"Author:",A_NORMAL);
			DrawStr(9+(offset?DrawStrWidth(puzzle->name,F_4x6):(puzzle->dim*4)),28,puzzle->author,A_NORMAL);
			DrawStr(9+(offset?DrawStrWidth(puzzle->name,F_4x6):(puzzle->dim*4)),40,"Completed in",A_NORMAL);
			
			seconds2str(1800-seconds,(char *)board_cleared);
			DrawStr(55+(offset?DrawStrWidth(puzzle->name,F_4x6):(puzzle->dim*4)),40,(char *)board_cleared,A_NORMAL);
			
			
			
			if((solved=add_best_time(puzzle->name,puzzle->author,1800-seconds))>0) {
				DrawStr(9+(offset?DrawStrWidth(puzzle->name,F_4x6):(puzzle->dim*4)),47,"New Record!",A_NORMAL);
			}
			else if(solved<0) {
				DrawStr(9+(offset?DrawStrWidth(puzzle->name,F_4x6):(puzzle->dim*4)),47,"Error adding record: Memory",A_NORMAL);
			}
			
			
			
		}
		else {
			DrawStr(0,0,"Game Over :(",A_NORMAL);
		}
		
		GKeyFlush();
		key=0;
		
		ST_busy(ST_NORMAL);
		
		while(key!=KEY_ENTER) {
			OSdequeue(&key,kbq);
		}
	}
	else {
		memcpy(buf,board_cleared,puzzle->dim*sizeof(unsigned short));
	}
	
	
	//indicate that no state was saved
	*(settings_file+21)=0;
	
	return 0;
}


inline short check_solved(PUZZLE *puzzle, unsigned short *check) {
	return !memcmp(puzzle->board,check,puzzle->dim*sizeof(unsigned short));
}


void calc_numbers(PUZZLE *puzzle) {
	short offset=puzzle->dim-1;
	short w;
	short h=0;
	short run=0; //how many pixels in a group
	short pos=0; //position in the numbers array
	short num=0; //number of groups
	char last=0; //last char seen
	
	
	puzzle->maxnumx=0;
	puzzle->maxnumy=0;
	
	
	for(w=offset;w>=0;w--) {
		for(last=0,num=0,run=0,h=0;h<puzzle->dim;h++) {
			if(puzzle->board[h]&(1<<w)) {
				run++;
				last=1;
			}
			else {
				if(last) {
					pos++;
					puzzle->numbers[pos]=run;
					num++;
				}
				
				run=0;
				last=0;
			}	
		}
		
		if(last) {
			pos++;
			puzzle->numbers[pos]=run;
			num++;
		}
		
		puzzle->numbers[pos-num]=num?num:1;
		
		if(!num) {
			puzzle->numbers[++pos]=0;
		}
		
		if(num>puzzle->maxnumy)
			puzzle->maxnumy=num;
		
		
		pos++;
			
	}
	
	if(!puzzle->maxnumy)
		puzzle->maxnumy=1;
	
	
	puzzle->maxnumy*=6;
	
	
	for(h=0;h<puzzle->dim;h++) {
		for(last=0,num=0,run=0,w=offset;w>=0;w--) {
			if(puzzle->board[h]&(1<<w)) {
				run++;
				last=1;
			}
			else {
				if(last) {
					pos++;
					puzzle->numbers[pos]=run;
					num++;
				}
				
				run=0;
				last=0;
			}	
		}
		
		if(last) {
			pos++;
			puzzle->numbers[pos]=run;
			num++;
		}
		
		puzzle->numbers[pos-num]=num?num:1;
		
		if(!num) {
			puzzle->numbers[++pos]=0;
		}
		
		
		if(num>puzzle->maxnumx)
			puzzle->maxnumx=num;
		
		
		pos++;
			
	}
	
	if(!puzzle->maxnumx)
		puzzle->maxnumx=1;
	
	puzzle->maxnumx*=5;
}


/*NUMBERS FORMAT: beginning with the horizontal row from left to right:
First, a number defining how many numbers there are for that row or column
Next, the numbers in order from top to bottom or left to right
Then the next column is defined.  After all the columns are defined, the rows are defined in the same way

*/

//pass NULL for board_marked to indicate edit mode (no numbers drawn)
void update_board(PUZZLE *puzzle, unsigned short *board_cleared, unsigned short *board_marked, short offset) {
	short i=0;
	short f;
	short y=puzzle->maxnumy;
	short x=puzzle->maxnumx;
	short j=0;
	short step=0;
	short num;
	short offset_fix=offset<0?0:offset;
	
	
	
	
	ScrRectFill(&(SCR_RECT){{0,y-1,x+(puzzle->dim*6),LCD_HEIGHT-1}},&(SCR_RECT){{0,y-1,x+(puzzle->dim*6),LCD_HEIGHT-1}},A_REVERSE);
	
	//update scrollbar
	if(((LCD_HEIGHT-puzzle->maxnumy)/6)-puzzle->dim<0) {
		ScrRectFill(&(SCR_RECT){{puzzle->maxnumx+(puzzle->dim*6)+4,puzzle->maxnumy+7,puzzle->maxnumx+(puzzle->dim*6)+6,LCD_HEIGHT-8}},&(SCR_RECT){{puzzle->maxnumx+(puzzle->dim*6)+4,puzzle->maxnumy+7,puzzle->maxnumx+(puzzle->dim*6)+6,LCD_HEIGHT-8}},A_REVERSE);
		DrawLine(puzzle->maxnumx+(puzzle->dim*6)+3,puzzle->maxnumy+7+((LCD_HEIGHT-8-(puzzle->maxnumy+7))*offset_fix)/abs((((LCD_HEIGHT-puzzle->maxnumy)/6)-puzzle->dim)),puzzle->maxnumx+(puzzle->dim*6)+7,puzzle->maxnumy+7+((LCD_HEIGHT-8-(puzzle->maxnumy+7))*offset_fix)/abs((((LCD_HEIGHT-puzzle->maxnumy)/6)-puzzle->dim)),A_NORMAL);
	}
	
	
	if(board_marked) {
		if(offset<0) {
			//draw numbers for columns
			for(i=0;i<puzzle->dim;i++,x+=6) {
				num=puzzle->numbers[j++];
				for(step=0,y=(puzzle->maxnumy)-30;step<num;j++,y+=6,step++) {
					if(puzzle->numbers[j]<10)
						DrawChar(x+2,((5-num)*6)+y,'0'+puzzle->numbers[j],A_NORMAL);
					else
						Sprite8(x+2,((5-num)*6)+y,5,numbers[puzzle->numbers[j]-10],LCD_MEM,SPRT_OR);
				}
			}
		}
		else {
			//cue up to row numbers
			for(f=0;f<puzzle->dim;f++) {
				j+=puzzle->numbers[j]+1;
			}
		}
		
		//add offset to row numbers
		for(f=0;f<offset_fix;f++) {
			j+=puzzle->numbers[j]+1;
		}
	}
	
	
	
	for(i=offset_fix,y=puzzle->maxnumy;i<puzzle->dim;i++,y+=6) {
		for(f=puzzle->dim-1,x=puzzle->maxnumx+1;f>=0;f--,x+=6) {
			if(board_marked && board_marked[i]&(1<<f))
				Sprite8(x+2,y+2,3,tile_marked,LCD_MEM,SPRT_OR);
			else if(board_cleared[i]&(1<<f))
				Sprite8(x+2,y+2,5,tile_cursor,LCD_MEM,SPRT_OR);
			
			Sprite8(x,y,7,tile_normal,LCD_MEM,SPRT_OR);
		}
		
		if(board_marked) {
			num=puzzle->numbers[j++];
			for(step=0,x=0;step<num;j++,x+=5,step++) {
				if(puzzle->numbers[j]<10)
					DrawChar((puzzle->maxnumx)-(num*5)+x,y+1,'0'+puzzle->numbers[j],A_NORMAL);
				else
					Sprite8((puzzle->maxnumx)-(num*5)+x,y+1,5,numbers[puzzle->numbers[j]-10],LCD_MEM,SPRT_OR);
			}
		}
	}

}
