 /*
    Peter's Picross - A remake for TI 68k-based calculators of the Nintendo Gameboy game, "Mario's Picross"
    Copyright (C) 2008  Peter Fernandes
    supersonicandtails@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    
    
    files.c: loading and writing files
*/

#include <tigcclib.h>
#include "main.h"
#include "engine.h"
#include "gfx.h"
#include "files.h"


char *settings_file;
HANDLE settings_handle;
char settings_file_isarchived;


short get_best_time(char *title, char *author) {
	short i=0;
	char *pos=settings_file+21+72;
	char num=*pos;
	
	
	pos++;
	
	
	for(i=0;i<num;i++) {
		if(!strcmp(title,pos) && !strcmp(author,pos+21)) {
			
			//read the record
			return *((short *)(pos+42));
		}
		
		pos+=44;
	}
	
	return -1;
}


short unarchive_settings(void) {
	
	if(settings_file_isarchived) {
		if(!EM_moveSymFromExtMem(SYMSTR("picross\\picset"),HS_NULL)) {
			return -1;
		}
		
		HeapUnlock(settings_handle);
		
		if(!(settings_file=open_file("picross\\picset",&settings_handle))) {
			EM_moveSymToExtMem(SYMSTR("picross\\picset"),HS_NULL);
			HeapLock(settings_handle);
			return -1;
		}
		HeapLock(settings_handle);
		
	}
	
	return 0;
}

short archive_settings(void) {
	
	
	if(!EM_moveSymToExtMem(SYMSTR("picross\\picset"),HS_NULL)) {
		return -1;
	}
	
	HeapUnlock(settings_handle);
	
	if(!(settings_file=open_file("picross\\picset",&settings_handle))) {
		EM_moveSymFromExtMem(SYMSTR("picross\\picset"),HS_NULL);
		HeapLock(settings_handle);
		return -1;
	}
	HeapLock(settings_handle);
	
	return 0;
}

short add_best_time(char *title, char *author, short seconds) {
	char *pos=settings_file+21+72;
	short i=0;
	char num=*pos;
	short err;
	
	pos++;
	
	
	if(unarchive_settings()) {
		return -1;
	}
	
	
	//attempt to update first
	for(i=0;i<num;i++) {
		if(!strcmp(title,pos) && !strcmp(author,pos+21)) {
			if(seconds<(*((short *)(pos+42)))) {
				//update the record
				*((short *)(pos+42))=seconds;
				
				err=1;
				goto end;
			}
			
			//a better record exists
			err=0;
			goto end;
		}
		
		
		pos+=44;
	}
	
	//create a new record
	
	HeapUnlock(settings_handle);
	
	if((settings_handle=HeapRealloc(settings_handle,2+21+72+1+7+(44*(num+1))))==H_NULL)
		return -2;
		
	HeapLock(settings_handle);
	
	
	//add two bytes because we don't care about the filesize header
	settings_file=HeapDeref(settings_handle)+2;
		
	
	*((short *)(settings_file-2))+=44;
	
	//increment num entry in file
	(*(settings_file+21+72))++;
	
	
	pos=settings_file+21+72+1+(44*num);
	
	//write the entry
	strcpy(pos,title);
	strcpy(pos+21,author);
	*((short *)(pos+42))=seconds;
	
	pos+=44;
	
	*(pos++)=0;
	*(pos++)='P';
	*(pos++)='I';
	*(pos++)='C';
	*(pos++)='S';
	*(pos++)=0;
	*(pos)=OTH_TAG;
	
	err=1;
	
	end:
	
	if(settings_file_isarchived) {
		archive_settings();
	}
	
	return err;
}



//returns a pointer to the file data
char *open_file(const char *filename, HANDLE *handle) {
	HANDLE localhandle;
	HSym symbol;
	SYM_ENTRY *symptr;
	char newstr[18];
	
	memset(newstr,0,sizeof(newstr));
	
	strcat(newstr+1,filename);
	
	if(!(symbol=SymFind(newstr+strlen(newstr+1)+1)).folder)
		return NULL;
	
	symptr=DerefSym(symbol);
	
	if(!(localhandle=symptr->handle))
		return NULL;
		
		
	if(handle) {
		*handle=localhandle;
		HeapLock(*handle);
	}
	
	
	if((!strcmp("picross\\picset",filename)) && symptr->flags.bits.archived==1) {
		settings_file_isarchived=1;
	}
	
		
	//add two bytes because we don't care about the filesize header
	return ((char *)HeapDeref(localhandle))+2;
}


short create_file(const char *filename, char **ptr, short size, HANDLE *handle) {
	HSym symbol;
	HANDLE localhandle;
	char *data;
	char newstr[18];
	
	
	memset(newstr,0,sizeof(newstr));
	
	memcpy(newstr+1,"picross\\",8);
	strcat(newstr+1+8,filename);
	
	
	//make sure the filename isn't already taken
	if(SymFind(newstr+strlen(newstr+1)+1).folder) {
		return -1;
	}

 
	
	if(!(symbol=SymAdd(newstr+strlen(newstr+1)+1)).folder) {
		return -2;
	}
	
	
	if(!(localhandle=HeapAlloc(size))) {
		return -3;
	}
	
	
	if(((DerefSym(symbol)->handle)=localhandle)==H_NULL) {
		return -4;
	}
	
	if(!(data=HeapDeref(localhandle))) {
		return -5;
	}
	
	
	if(handle) {
		*handle=localhandle;
		HeapLock(*handle);
	}
	
	*ptr=data;
	
	return 0;
}


short write_puzzle_file(const char *filename, PUZZLE *puzzle) {
	char *data;
	short err;
	
	if((err=create_file(filename,&data,9+48+(puzzle->dim*sizeof(unsigned short)),NULL))) {
		return err;
	}
	
	
	//file size
	*(((unsigned short *)data)++)=7+48+(puzzle->dim*sizeof(unsigned short));
	
	*(((unsigned long *)data)++)=0x70706200;
	strcpy(data,puzzle->name);
	data+=21;
	strcpy(data,puzzle->author);
	data+=21;
	*data=puzzle->dim;
	
	//pad with one byte to word-align the board data
	data+=2;
	
	memcpy(data,puzzle->board,puzzle->dim*sizeof(unsigned short));
	data+=puzzle->dim*sizeof(unsigned short);
	
	*(data++)=0;
	*(data++)='P';
	*(data++)='I';
	*(data++)='C';
	*(data++)='B';
	*(data++)=0;
	*(data)=OTH_TAG;
	
	
	
	return 0;
}


int dialog_load_puzzle(PUZZLE *puzzle, short autochoose, char *filename) {
	SYM_ENTRY *cursym;
	short size=2;
	char name[18];
	char dims[6];
	PUZZLE_ENTRY *file;
	char *data;
	unsigned short key=0;
	short offset=0;
	short update=1;
	short y=0;
	short oldpos=0;
	short pos=0;
	short best;
	short restore_entry=-1;
	
	
	
	
	if(!autochoose && !filename) {
		ClrScr();
		
		DrawStr((LCD_WIDTH-DrawStrWidth("Load Puzzle",F_4x6))/2,0,"Load Puzzle",A_NORMAL);
		
		FontSetSys(F_4x6);
		DrawStr(0,20,"Name",A_NORMAL);
		DrawStr(45,20,"Best",A_NORMAL);
		DrawStr(70,20,"Author",A_NORMAL);
		DrawStr(140,20,"Dim",A_NORMAL);
	}
	
	
	cursym=SymFindFirst(SYMSTR("picross"),FO_SINGLE_FOLDER);
	
	if(!cursym) {
		return -2;
	}
	
	if(!(file=malloc(sizeof(PUZZLE_ENTRY)))) {
		return -4;
	}
	
	memcpy(name,"picross\\",8);
	
	

 	for(size=2;cursym;) {
		memset(name+8,0,sizeof(name)-8);
		memcpy(name+8,cursym->name,8);
		
		
 		if(!(data=open_file(name,NULL))) {
 			return -3;
 		}
 		
 		
 		if(*((unsigned long *)data)==0x70706200) { //it's most likely a picross file
 			
 			if(!(file=realloc(file,size*sizeof(PUZZLE_ENTRY)))) {
				return -4;
			}
				
			strcpy(file[size-2].name,name+8);
			
			file[size-2].data=data;
 			
 			if(!strcmp(filename,file[size-2].name))
 				restore_entry=size-2;
 			
 			size++;
 		}
 		
 		cursym=SymFindNext();
 	}
	
	
	if(size<=2) {
		return -2;
	}
	
	if(autochoose) {
		/*if(size<3)
			return -2;*/
		
		y=random(size-3);
		goto end;
	}
	else if(filename) {
		/*if(size<3)
			return -2;*/
			
		if(restore_entry>=0)
			y=restore_entry;
		else
			return -3;
		
		goto end;
		
	}
	
	
	pos=0;
	oldpos=0;
	key=1;
	
	
	
	
	do {
		
		if((key&~KB_AUTOREPEAT)==(unsigned short)KEY_UP+KEY_DIAMOND) { //diamond+up
			oldpos=pos;
			goto top_of_list;
		}
		else if((key&~KB_AUTOREPEAT)==(unsigned short)KEY_DOWN+KEY_DIAMOND) { //diamond+down
			oldpos=pos;
			goto end_of_list;
		}
		else if((key&~KB_AUTOREPEAT)==KEY_UP+KEY_2ND) { //2ND+up	
			oldpos=pos;
			
			if(offset)
				update=1;
			
			pos=0;
			y-=(y>=SLOTS-1?(SLOTS-1):y);
			
			offset=y;
			
			key=0;
		}
		else if((key&~KB_AUTOREPEAT)==KEY_DOWN+KEY_2ND) { //2ND+down
			oldpos=pos;
			
			if(offset<=size-3-SLOTS)
				update=1;
			
			pos=(size-2)<SLOTS?(size-3):SLOTS-1;
			
			y+=(y<(size-2-SLOTS)?SLOTS:(size-3)-y);
			
			if(y>=SLOTS)
				offset=y-SLOTS+1;
			
			key=0;
		}
		else if((key&~KB_AUTOREPEAT)==KEY_UP) { //Up Arrow
			oldpos=pos;
				
			if(y) {
				if(!(y-offset)) {
					offset--;
					update=1;
				}
				else
					pos--;
				
				
				y--;
			}
			else {
				end_of_list:
				
				offset=(size-2)-SLOTS>0 ? (size-2)-SLOTS : 0;
				y=size-3;
				pos=(size-2)<SLOTS-1?(size-3):SLOTS-1;
				if(offset)
					update=1;
			}
		}
		else if((key&~KB_AUTOREPEAT)==KEY_DOWN) { //Down Arrow
			oldpos=pos;
			
			if(y<size-3) {
				if(y+1-offset==SLOTS) {
					offset++;
					update=1;
				}
				else
					pos++;
					
				y++;
			}
			else {
				top_of_list:
				
				if(offset)
					update=1;
				y=0;
				offset=0;
				pos=0;
			}
		}
		else if((key&~KB_AUTOREPEAT)==KEY_ESC) { //ESC Key
			free(file);
			return -5;
		}
		else if(key!=1) {
			goto bottom;
		}
		
		
		if(update) {
			ScrRectFill(&(SCR_RECT){{0,30,LCD_WIDTH-1,LCD_HEIGHT-1}},&(SCR_RECT){{0,30,LCD_WIDTH-1,LCD_HEIGHT-1}},A_REVERSE);
			
			if(size-2-offset>SLOTS) {
				DrawStr(0,LCD_HEIGHT-7,"...",A_NORMAL);
			}
			
			//update list
			for(key=offset;(signed short)(key-offset)<SLOTS && (signed short)key<size-2;key++) {
				//puzzle name
				DrawStr(0,((key-offset)*7)+30,file[key].name,A_NORMAL);
				
				/*
				//Draw an elipses if the string can't fit in the column
				if(DrawStrWidth(file[key].puzzle.name,F_4x6)>45) {
					DrawStr(63,((key-offset)*7)+30,"... ",A_REPLACE);
				}*/
				
				//best time
				best=get_best_time(file[key].data+OFFSET_NAME,file[key].data+OFFSET_AUTHOR);
				
				if(best>0) {
					seconds2str(best,dims);
				}
				else {
					strcpy(dims,"----");
				}
				
				DrawStr(45,((key-offset)*7)+30,dims,A_NORMAL);
				
				//puzzle author
				DrawStr(70,(((key-offset))*7)+30,file[key].data+OFFSET_AUTHOR,A_REPLACE);
				
				/*
				//Draw an elipses if the string can't fit in the column
				if(DrawStrWidth(file[key].puzzle.data+OFFSET_AUTHOR,F_4x6)>65) {
					DrawStr(132,((key-offset)*7)+30,"... ",A_REPLACE);
				}*/
				
				//puzzle dimensions
				sprintf(dims,"%dx%d",*(file[key].data+OFFSET_DIM),*(file[key].data+OFFSET_DIM));
				DrawStr(140,((key-offset)*7)+30,dims,A_REPLACE);
			}
			
			update=0;
		}
		else {
			ScrRectFill(&(SCR_RECT){{0,(oldpos*7)+30,69,(oldpos*7)+35}},&(SCR_RECT){{0,(oldpos*7)+30,69,(oldpos*7)+35}},A_XOR);
		}
		
		ScrRectFill(&(SCR_RECT){{0,(pos*7)+30,69,(pos*7)+35}},&(SCR_RECT){{0,(pos*7)+30,69,(pos*7)+35}},A_XOR);
		
		bottom:
		
	} while((key=ngetchx())!=KEY_ENTER);
	
	
	end:
	
	puzzle->name=file[y].data+OFFSET_NAME;
	puzzle->author=file[y].data+OFFSET_AUTHOR;
	puzzle->dim=*(file[y].data+OFFSET_DIM);
	puzzle->board=(unsigned short *)(file[y].data+OFFSET_DATA);
	memcpy(puzzle->filename,file[y].name,8);
	
	
	for(key=0;(signed)key<puzzle->dim;key++) {
		if(puzzle->board[key])
			break;
	}
	
	
	
	free(file);
	
	
	
	//return error if the puzzle is empty
	if((signed)key==puzzle->dim)
		return -6;
	
	
	
	
	return 0;
}



